import datetime
import importlib
import inspect
import json
import os
import requests
import sys
import time
from pprint import pprint

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.core.mail import mail_admins
from django.core.management.base import BaseCommand, CommandError
from django.db.models import F
from django.utils import timezone

from saleboxdjango.lib.common import update_natural_sort
from saleboxdjango.lib.translation import get_translations
from saleboxdjango.models import *


class Command(BaseCommand):
    SOFTWARE_VERSION = '0.1.6'

    def handle(self, *args, **options):
        now = time.time()

        # bail out in sync already running
        sync_start = int(self.timer_get('saleboxsync_sync_start'))
        if (time.time() - sync_start) < 120:
            print('sync in progress: bailing out!')
            return

        # step #1: if a user needs to be converted to a member, queue it
        self.timer_set('saleboxsync_sync_start', time.time())
        user_equals_member = settings.SALEBOX['MEMBER']['USER_EQUALS_MEMBER']
        if user_equals_member != 'MANUAL':
            # 'ALWAYS': every website user should be a salebox member
            if user_equals_member == 'ALWAYS':
                user_ids = get_user_model() \
                            .objects \
                            .filter(salebox_member_id__isnull=True) \
                            .values_list('id', flat=True)

            # 'PURCHASE': every website user that makes a purchase should
            # be a salebox member
            if user_equals_member == 'PURCHASE':
                user_ids = CheckoutStore \
                            .objects \
                            .filter(status__lte=30) \
                            .filter(user__isnull=False) \
                            .values_list('user', flat=True)
                if len(user_ids) > 0:
                    user_ids = get_user_model() \
                                .objects \
                                .filter(id__in=list(set(user_ids))) \
                                .filter(salebox_member_id__isnull=True) \
                                .values_list('id', flat=True)

            # give those users a salebox_member_id and details to sync
            if len(user_ids) > 0:
                users = get_user_model().objects.filter(id__in=user_ids)
                for u in users:
                    u.set_salebox_member_sync('email', u.email)
                    # potential TODO:
                    # have a list of key mappings in the config
                    # which automatically get added to the sync
                    # JSON at this point

        # step #2: push users with member updates
        self.timer_set('saleboxsync_sync_start', time.time())
        self.push_members()

        # step #3: pull data
        # members must be up-to-date before POSTing transactions
        pull_start = int(self.timer_get('saleboxsync_pull_start'))
        # pull_start = 0
        if (time.time() - pull_start) > 599:
            self.pull_loop()

        # step 4: POST transactions
        transactions_pushed = self.push_transactions()
        if transactions_pushed:
            # pull again to get the event emails
            time.sleep(2)
            self.pull_loop()

        # step 5: run event handler
        self.timer_set('saleboxsync_sync_start', time.time())
        if 'EVENT_HANDLER' in settings.SALEBOX:
            sys.path.insert(0, settings.SALEBOX['EVENT_HANDLER']['APP_FOLDER'])
            import_path = settings.SALEBOX['EVENT_HANDLER']['CLASS']
            path, name = import_path.rsplit('.', 1)
            c = getattr(importlib.import_module(path), name)
            c()

        # step 6: send django-mail-queue emails
        try:
            self.timer_set('saleboxsync_sync_start', time.time())
            from mailqueue.models import MailerMessage
            MailerMessage.objects.send_queued(limit=30)
        except:
            pass

        # step 7: update stock levels
        self.pull_inventory()

        # step 8: pull latest member transaction history
        self.pull_transaction_history()

        # step 9: check for timed-out checkout stores
        css = CheckoutStore.objects.filter(status__lt=30)
        for cs in css:
            cs.check_for_timeout()

        # finished: reset the clock
        self.timer_set('saleboxsync_sync_start', 0.0)
        print('Finished in %.3fs' % (time.time() - now))

    def init_post(self):
        return {
            'pos': settings.SALEBOX['API']['KEY'],
            'license': settings.SALEBOX['API']['LICENSE'],
            'platform_type': 'ECOMMERCE',
            'platform_version': self.SOFTWARE_VERSION
        }

    def pull(self):
        post = self.init_post()

        # populate last updates
        sync_from_dict = self.pull_get_sync_from_dict()
        for code in sync_from_dict:
            post['lu_%s' % code] = sync_from_dict[code]

        print()
        print('Attempting sync with the following parameters:')
        pprint(sync_from_dict)
        print()

        # do request
        url = '%s/api/pos/v2/sync' % settings.SALEBOX['API']['URL']
        try:
            r = requests.post(url, data=post)
            have_response = True
        except:
            self.send_admin_email('Could not connect to Salebox POSv2 API')
            have_response = False

        try:
            if have_response:
                response = r.json()
                # pprint(response)

                # display debug meta
                print('Output received (\'data\' attribute truncated to len(data)):')
                response_meta = r.json()

                # exit if connection error
                if response_meta['status'] != 'OK':
                    print('Connection error: %s' % response_meta['status'])
                    return

                # display meta object
                for code in response_meta['sync']:
                    code['data'] = len(code['data'])
                pprint(response_meta)
                print()
                print('Updating:')

                # save data
                for i, value in enumerate(response['sync']):
                    if value['code'] == 'attribute':
                        self.pull_model_attribute(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'attribute_item':
                        self.pull_model_attribute_item(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'country':
                        self.pull_model_country(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'country_state':
                        self.pull_model_country_state(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'discount_group':
                        self.pull_model_discount_group(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'discount_ruleset':
                        self.pull_model_discount_ruleset(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'event':
                        self.pull_model_event(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'i18n':
                        self.pull_model_translation(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'member':
                        self.pull_model_member(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'member_group':
                        self.pull_model_member_group(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'product':
                        self.pull_model_product(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'product_category':
                        self.pull_model_product_category(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'product_variant':
                        self.pull_model_product_variant(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    elif value['code'] == 'product_variant_image':
                        self.pull_model_product_variant_image(
                            value['data'],
                            value['lu'],
                            sync_from_dict
                        )

                    else:
                        print('Error: %s' % value['code'])

                return response['resync_now']
                # return False
        except:
            self.send_admin_email()
            return False

    def pull_get_sync_from_dict(self):
        lus = LastUpdate.objects.all()

        # populate dict
        lu = {}
        for l in lus:
            lu[l.code] = l.value

        # fill in the gaps
        for code in [
            'attribute',
            'attribute_item',
            'country',
            'country_state',
            # 'discount_group',
            # 'discount_ruleset',
            'event',
            'i18n',
            'member',
            'member_group',
            'product',
            'product_category',
            'product_variant',
            'product_variant_image',
        ]:
            if code not in lu:
                LastUpdate(code=code, value=0.0).save()
                lu[code] = 0.0

        return lu

    def pull_image(self, id, urlpath, dir):
        self.timer_set('saleboxsync_sync_start', time.time())
        self.timer_set('saleboxsync_pull_start', time.time())

        # ensure target dir exists
        target_dir = '%s/salebox/%s' % (settings.MEDIA_ROOT, dir)
        try:
            os.makedirs(target_dir)
        except:
            pass

        # fetch image
        try:
            url = '%s/%s' % (settings.SALEBOX['API']['URL'], urlpath)
            suffix = urlpath.split('.')[-1]
            filename = '%s.%s.%s' % (id, urlpath.split('/')[-1][0:6], suffix)
            target = '%s/%s' % (target_dir, filename)
            r = requests.get(url)
            if r.status_code == 200:
                open(target, 'wb').write(r.content)
                return filename, True
        except:
            self.send_admin_email()

        return None, False

    def pull_images(self):
        # product category
        imgs = ProductCategory \
                .objects \
                .exclude(image__isnull=True) \
                .filter(local_image__isnull=True) \
                .filter(active_flag=True)
        for img in imgs:
            path, success = self.pull_image(img.id, img.image[1:], 'pospc')
            if success:
                print('.', end='')
                img.local_image = path
                img.save()

        # product
        imgs = Product \
                .objects \
                .exclude(image__isnull=True) \
                .filter(local_image__isnull=True) \
                .filter(active_flag=True)
        for img in imgs:
            path, success = self.pull_image(img.id, img.image[1:], 'posp')
            if success:
                print('.', end='')
                img.local_image = path
                img.save()

        # product variant
        imgs = ProductVariant \
                .objects \
                .exclude(image__isnull=True) \
                .filter(local_image__isnull=True) \
                .filter(active_flag=True)
        for img in imgs:
            path, success = self.pull_image(img.id, img.image[1:], 'pospv')
            if success:
                print('.', end='')
                img.local_image = path
                img.save()

        # product variant image
        imgs = ProductVariantImage \
                .objects \
                .exclude(img__isnull=True) \
                .filter(local_img__isnull=True) \
                .filter(active_flag=True)
        for img in imgs:
            path, success = self.pull_image(img.id, img.img, 'pvi')
            if success:
                print('.', end='')
                img.local_img = path
                img.save()

    def pull_inventory(self):
        url = '%s/api/pos/v2/inventory' % settings.SALEBOX['API']['URL']
        post = self.init_post()
        inventory = []

        # attempt sync 'all'
        sync_all = int(self.timer_get('saleboxsync_inventory_all'))
        # sync_all = 0
        if (time.time() - sync_all) > (60 * 60 * 24):
            print('Inventory sync all')
            post['request'] = 'all'
        else:
            # attempt sync 'recent'
            sync_recent = int(self.timer_get('saleboxsync_inventory_recent'))
            # sync_recent = 999
            if (time.time() - sync_recent) > (60 * 15):
                print('Inventory sync recent')
                post['request'] = 'recent'
                post['time_offset'] = (60 * 15)
            else:
                # attempt sync 'lowstock'
                variant_ids = ProductVariant \
                                .objects \
                                .filter(active_flag=True) \
                                .filter(product__active_flag=True) \
                                .filter(product__category__active_flag=True) \
                                .filter(available_on_ecom=True) \
                                .filter(stock_total__gt=0) \
                                .filter(stock_total__lte=F('ecommerce_low_stock_threshold')) \
                                .values_list('id', flat=True)
                if len(variant_ids) > 0:
                    print('Inventory sync low-stock')
                    post['request'] = 'lowstock'
                    post['variant_ids'] = ','.join([str(i) for i in variant_ids])

        # bail if nothing to fetch
        if 'request' not in post:
            return

        # fetch from server
        try:
            r = requests.post(url, data=post)
            if r.status_code == 200:
                result = r.json()
                if result['status'] == 'OK':
                    if 'inventory' in result:
                        inventory = result['inventory']
                    else:
                        self.send_admin_email('API response recieved but inventory missing (request type: %s)' % post['request'])
                else:
                    self.send_admin_email('API response recieved but status not OK (request type: %s)' % post['request'])
            else:
                self.send_admin_email('API response status code: %s (request type: %s)' % (r.status_code, post['request']))
        except:
            self.send_admin_email('Unknown error connecting to Salebox POSv2 API (request type: %s)' % post['request'])

        # apply results
        if len(inventory) > 0:
            for i in inventory:
                variant_id = i[0]
                stock_count = i[1]
                pv = ProductVariant.objects.filter(id=variant_id).first()
                if pv is not None:
                    if pv.stock_count != stock_count:
                        pv.stock_count = stock_count
                        pv.save()

            # set timers
            if post['request'] == 'all':
                self.timer_set('saleboxsync_inventory_all', time.time())
                self.timer_set('saleboxsync_inventory_recent', time.time())
            elif post['request'] == 'recent':
                self.timer_set('saleboxsync_inventory_recent', time.time())

    def pull_loop(self):
        while True:
            self.timer_set('saleboxsync_sync_start', time.time())
            self.timer_set('saleboxsync_pull_start', time.time())
            pull_result = self.pull()
            if pull_result:
                print()
                print('sleeping...')
                print()
                time.sleep(5)
            else:
                break

        # pull images
        print()
        print('Downloading images...')
        self.pull_images()
        print()

    def pull_model_attribute(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Attribute.objects.get_or_create(id=d['id'])
                o.code = d['code']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'attribute',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x Attribute' % len(data))
        except:
            self.send_admin_email()

    def pull_model_attribute_item(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = AttributeItem.objects.get_or_create(id=d['id'])
                o.attribute = Attribute.objects.get(id=d['attribute'])
                o.slug = d['slug']
                o.value = d['value']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'attribute_item',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x AttributeItem' % len(data))
        except:
            self.send_admin_email()

    def pull_model_country(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Country.objects.get_or_create(id=d['id'])
                o.code = d['code']
                o.default = d['default']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'country',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x Country' % len(data))
        except:
            self.send_admin_email()

    def pull_model_country_state(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = CountryState.objects.get_or_create(id=d['id'])
                o.country = Country.objects.get(id=d['country'])
                o.code = d['code']
                o.full_code = d['full_code']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'country_state',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x CountryState' % len(data))
        except:
            self.send_admin_email()

    def pull_model_discount_group(self, data, api_lu, sync_from_dict):
        # for d in data
        #
        #

        # update sync_from
        """
        self.pull_set_sync_from_dict(
            'discount_group',
            len(data) < 100,
            sync_from_dict,
            api_lu
        )
        """

    def pull_model_discount_ruleset(self, data, api_lu, sync_from_dict):
        # for d in data
        #
        #

        # update sync_from
        """
        self.pull_set_sync_from_dict(
            'discount_ruleset',
            len(data) < 100,
            sync_from_dict,
            api_lu
        )
        """

    def pull_model_event(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Event.objects.get_or_create(id=d['id'])
                o.event = d['event']
                o.salebox_member_id = d['salebox_member_id']
                o.transaction_guid = d['transaction_guid']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'event',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x Event' % len(data))
        except:
            self.send_admin_email()

    def pull_model_member(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                # retrieve lookups
                group = MemberGroup.objects.get(id=d['group'])
                parent = Member.objects.get_or_create(id=d['id'])
                try:
                    gwc = MemberGroup.objects.get(id=d['group_when_created'])
                except:
                    gwc = None

                # get parent
                parent = None
                if d['parent'] is not None:
                    parent, created = Member.objects.get_or_create(id=d['parent'])

                # country / states
                country = None
                country_state = None
                if d['country'] is not None:
                    country = Country.objects.get(id=d['country'])
                if d['country_state'] is not None:
                    country_state = CountryState.objects.get(id=d['country_state'])

                # create object
                o, created = Member.objects.get_or_create(id=d['id'])
                o.group = group
                o.parent = parent
                o.group_when_created = gwc
                o.country = country
                o.country_state = country_state

                # update
                for a in [
                    'active_flag',
                    'address_1',
                    'address_2',
                    'address_3',
                    'address_4',
                    'address_5',
                    'boolean_1',
                    'boolean_2',
                    'boolean_3',
                    'boolean_4',
                    'boolean_5',
                    'boolean_6',
                    'date_of_birth',
                    'email',
                    'gender',
                    'guid',
                    'id_card',
                    'join_date',
                    'name_first',
                    'name_last',
                    'phone_1',
                    'phone_2',
                    'postcode',
                    'salebox_member_id',
                    'status',
                    'string_1',
                    'string_2',
                    'string_3',
                    'string_4',
                    'string_5',
                    'string_6',
                    'string_7',
                    'string_8',
                    'string_9',
                    'string_10',
                    'string_11',
                    'string_12',
                    'string_13',
                    'string_14',
                    'title',
                ]:
                    setattr(o, a, d[a])

                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'member',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x Member' % len(data))
        except:
            self.send_admin_email()

    def pull_model_member_group(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = MemberGroup.objects.get_or_create(id=d['id'])
                o.name = d['name']
                o.flat_discount_percentage = d['flat_discount_percentage']
                o.can_be_parent = d['can_be_parent']
                o.default_group = d['default_group']
                o.active_flag = d['active_flag']
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'member_group',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x MemberGroup' % len(data))
        except:
            self.send_admin_email()

    def pull_model_product(self, data, api_lu, sync_from_dict):
        def sync_attributes(product, attribute_number, id_list):
            attributes_changed = False
            attribute_m2m = getattr(product, 'attribute_%s' % attribute_number)

            # remove
            for a in attribute_m2m.all():
                if a.id not in id_list:
                    attribute_m2m.remove(a)
                    attributes_changed = True

            # add
            for id in id_list:
                ai = AttributeItem.objects.get(id=id)
                attribute_m2m.add(ai)
                attributes_changed = True

            # save
            if attributes_changed:
                product.save()

        try:
            for d in data:
                # create object
                o, created = Product.objects.get_or_create(id=d['id'])
                o.category = ProductCategory.objects.get(id=d['category'])

                # do we need to reset the local_image flag?
                if not created and o.image != d['image']:
                    o.local_image = None

                # update
                for a in [
                    'active_flag',
                    'bestseller_rank',
                    'boolean_1',
                    'boolean_2',
                    'boolean_3',
                    'boolean_4',
                    'i18n',
                    'image',
                    'inventory_flag',
                    'name',
                    'season',
                    'sold_by',
                    'slug',
                    'string_1',
                    'string_2',
                    'string_3',
                    'string_4',
                    'vat_applicable',
                ]:
                    setattr(o, a, d[a])

                o.save()

                # sync attributes
                for i in range(1, 11):
                    sync_attributes(o, i, d['attribute_%s' % i])

            # update sync_from
            self.pull_set_sync_from_dict(
                'product',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x Product' % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_category(self, data, api_lu, sync_from_dict):
        try:
            # pass #1: save with parent = None to avoid mptt errors
            for d in data:
                o, created = ProductCategory.objects.get_or_create(id=d['id'])

                # do we need to reset the local_image flag?
                if not created and o.image != d['image']:
                    o.local_image = None

                o.parent = None
                o.active_flag = d['active_flag']
                o.boolean_1 = d['boolean_1']
                o.boolean_2 = d['boolean_2']
                o.boolean_3 = d['boolean_3']
                o.boolean_4 = d['boolean_4']
                o.i18n = d['i18n']
                o.image = d['image']
                o.name = d['name']
                o.seasonal_flag = d['seasonal_flag']
                o.slug = d['slug']
                o.short_name = d['short_name']
                setattr(o, o._mptt_meta.left_attr, d['mptt_left'])
                setattr(o, o._mptt_meta.level_attr, d['mptt_level'])
                setattr(o, o._mptt_meta.parent_attr, None)
                setattr(o, o._mptt_meta.right_attr, d['mptt_right'])
                setattr(o, o._mptt_meta.tree_id_attr, d['mptt_tree_id'])
                o.save()

            # pass #2: re-save the categories with parents (where applicable)
            for d in data:
                parent = None
                if d['parent'] is not None:
                    parent = ProductCategory.objects.get(id=d['parent'])

                o = ProductCategory.objects.get(id=d['id'])
                o.parent = parent
                setattr(o, o._mptt_meta.parent_attr, parent)
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'product_category',
                True,
                sync_from_dict,
                api_lu
            )

            print('%s x ProductCategory' % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_variant(self, data, api_lu, sync_from_dict):
        def sync_attributes(variant, attribute_number, id_list):
            attributes_changed = False
            attribute_m2m = getattr(variant, 'attribute_%s' % attribute_number)

            # remove
            for a in attribute_m2m.all():
                if a.id not in id_list:
                    attribute_m2m.remove(a)
                    attributes_changed = True

            # add
            for id in id_list:
                ai = AttributeItem.objects.get(id=id)
                attribute_m2m.add(ai)
                attributes_changed = True

            # save
            if attributes_changed:
                variant.save()

        try:
            for d in data:
                # create object
                o, created = ProductVariant.objects.get_or_create(id=d['id'])
                o.product = Product.objects.get(id=d['product'])

                # do we need to reset the local_image flag?
                if not created and o.image != d['image']:
                    o.local_image = None

                # update
                for a in [
                    'active_flag',
                    'available_to_order',
                    'available_on_pos',
                    'available_on_ecom',
                    'bestseller_rank',
                    'barcode',
                    'bo_name',
                    'boolean_1',
                    'boolean_2',
                    'boolean_3',
                    'boolean_4',
                    'color',
                    'date_1',
                    'date_2',
                    'ecommerce_description',
                    'ecommerce_low_stock_threshold',
                    'i18n',
                    'image',
                    'int_1',
                    'int_2',
                    'int_3',
                    'int_4',
                    'loyalty_points',
                    'member_discount_applicable',
                    'name',
                    'plu',
                    'preorder_flag',
                    'price',
                    'sale_percent',
                    'sale_price',
                    'shelf_expiry_type',
                    'shelf_life_days',
                    'shipping_weight',
                    'size',
                    'size_order',
                    'size_uom',
                    'sku',
                    'slug',
                    'string_1',
                    'string_2',
                    'string_3',
                    'string_4',
                    'warehouse_location',
                ]:
                    setattr(o, a, d[a])

                o.save()

                # sync attributes
                for i in range(1, 11):
                    sync_attributes(o, i, d['attribute_%s' % i])

            # update sync_from
            self.pull_set_sync_from_dict(
                'product_variant',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            # update the natural sort value
            if len(data) > 0:
                update_natural_sort(ProductVariant, 'name', 'name_sorted')

            print('%s x ProductVariant' % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_variant_image(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                # create object
                o, created = ProductVariantImage.objects.get_or_create(
                    id=d['id'],
                    variant=ProductVariant.objects.get(id=d['variant'])
                )

                # update
                for a in [
                    'active_flag',
                    'description',
                    'img',
                    'img_height',
                    'img_width',
                    'order',
                    'title'
                ]:
                    setattr(o, a, d[a])

                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                'product_variant_image',
                len(data) < 100,
                sync_from_dict,
                api_lu
            )

            print('%s x ProductVariantImage' % len(data))
        except:
            self.send_admin_email()

    def pull_model_translation(self, data, api_lu, sync_from_dict):
        try:
            if 'translations' in data:
                for language_code in data['translations']:
                    self.pull_model_translation_recurse(language_code, data['translations'][language_code])

            # update sync_from
            self.pull_set_sync_from_dict(
                'i18n',
                True,
                sync_from_dict,
                api_lu
            )

            # rebuild cache
            get_translations(rebuild=True)

            print('%s x Translation' % len(data))
        except:
            self.send_admin_email()

    def pull_model_translation_recurse(self, language_code, data, prefix=''):
        for d in data:
            if isinstance(data[d], str):
                t, created = Translation.objects.get_or_create(
                    language_code=language_code,
                    key='%s%s' % (prefix, d)
                )
                t.value = data[d]
                t.save()
            else:
                self.pull_model_translation_recurse(language_code, data[d], '%s.' % d)

    def pull_set_sync_from_dict(self, code, increment, sync_from_dict, api_lu):
        if increment:
            if timezone.now().timestamp() - api_lu > 20:
                if api_lu == 0:
                    api_lu = 1.0
                else:
                    api_lu += 0.00001

        if api_lu > sync_from_dict[code]:
            lu = LastUpdate.objects.get(code=code)
            lu.value = float(api_lu)
            lu.save()
            cache.clear()

    def pull_transaction_history(self):
        members = Member \
                    .objects \
                    .filter(salebox_transactionhistory_request_dt__isnull=False) \
                    .order_by('salebox_transactionhistory_request_dt')[0:1000]

        for m in members:
            self.timer_set('saleboxsync_sync_start', time.time())
            url = '%s/api/pos/v2/member-transaction-history' % (
                settings.SALEBOX['API']['URL'],
            )
            post = self.init_post()
            post['salebox_member_id'] = str(m.salebox_member_id)

            # fetch from server
            try:
                r = requests.post(url, data=post)
                result = r.json()
                if 'transactions' in result:
                    m.transactionhistory_update_data(result['transactions'])
            except:
                self.send_admin_email('Could not connect to Salebox POSv2 API')

    def push_member(self, user):
        self.timer_set('saleboxsync_sync_start', time.time())

        # define a salebox_member_id if one doesn't exist
        if user.salebox_member_id is None:
            user.create_salebox_member_id()

        # build post
        post = self.init_post()
        post['salebox_member_id'] = user.salebox_member_id
        post['salebox_member_sync'] = json.dumps(user.salebox_member_sync)

        # do request
        url = '%s/api/pos/v2/member-update' % settings.SALEBOX['API']['URL']
        try:
            r = requests.post(url, data=post)
            if r.status_code == 200:
                user.salebox_member_sync = None
                user.save()
                self.timer_set('saleboxsync_pull_start', 0.0)
        except:
            self.send_admin_email('Failed to POST member data')

    def push_members(self):
        users = get_user_model() \
                    .objects \
                    .filter(salebox_member_sync__isnull=False)
        for user in users:
            self.push_member(user)

    def push_transaction(self, store):
        # build POST
        data = self.init_post()
        data['transaction'] = json.dumps({
            'basket': self.transaction_basket(store),
            'customer': self.transaction_customer(store),
            'extras': self.transaction_extras(store),
            'invoice': self.transaction_invoice(store),
            'manual_discount': None,
            'member': self.transaction_member(store),
            'meta': self.transaction_meta(store),
            'payment': self.transaction_payment(store),
            'shipping': self.transaction_shipping(store),
            'stored_value_load': None,
            'total': self.transaction_total(store)
        })

        # send
        url = '%s/api/pos/v2/transaction/create' % settings.SALEBOX['API']['URL']
        try:
            r = requests.post(url, data=data)
            have_response = True
        except:
            self.send_admin_email()
            have_response = False

        # check response
        if have_response:
            if r.json()['status'] == 'OK':
                store.status = 31
                store.save()

    def push_transactions(self):
        transactions_pushed = False

        css = CheckoutStore.objects.filter(status=30)
        for cs in css:
            self.timer_set('saleboxsync_sync_start', time.time())

            # ensure this user has a corresponding salebox_member
            user_equals_member = settings.SALEBOX['MEMBER']['USER_EQUALS_MEMBER']
            if user_equals_member != 'MANUAL' and cs.user is not None:
                user = get_user_model().objects.get(id=cs.user)
                member = Member.objects.filter(salebox_member_id=user.salebox_member_id).first()
                if member is None:
                    # skip posting this transaction for now...
                    # get on the next time round
                    if user.salebox_member_sync is None:
                        user.set_salebox_member_sync('email', user.email)
                    continue

            # push transaction
            self.push_transaction(cs)
            transactions_pushed = True

        return transactions_pushed

    def send_admin_email(self, extra=None):
        caller = inspect.currentframe().f_back
        msg = [
            'saleboxsync error occurred:',
            '',
            '         Website: %s' % settings.ALLOWED_HOSTS[0],
            '      Backoffice: %s' % settings.SALEBOX['API']['URL'],
            '  Error function: %s' % caller.f_code.co_name,
            '    Error lineno: %s' % caller.f_lineno,
        ]

        if extra is not None:
            msg.append('           Notes: %s' % extra)

        if settings.DEBUG:
            print('--------------------------------------------------------------------------')
            print('\n'.join(msg))
            print('--------------------------------------------------------------------------')
        else:
            mail_admins(
                'saleboxsync error: %s' % settings.ALLOWED_HOSTS[0],
                '\n'.join(msg)
            )

    def timer_get(self, code, default_value=0.0):
        try:
            return LastUpdate.objects.get(code=code).value
        except:
            LastUpdate(code=code, value=default_value).save()
            return default_value

    def timer_set(self, code, value):
        o = LastUpdate.objects.filter(code=code).first()
        if o is None:
            o = LastUpdate(code=code)
        o.value = value
        o.save()

    def transaction_basket(self, store):
        basket = []
        for b in store.data['basket']['items']:
            v = b['variant']

            basket.append({
                'discount_ruleset': None,
                'qty_loyalty': v.get('qty_loyalty', None),
                'price_original': v['qty_price'],
                'price_modified': v['qty_sale_price'],
                'product_weight': None,
                'quantity': b['qty'],
                'shipping_weight': v['shipping_weight'],
                'unit_price': v['price'],
                'variant_id': v['id'],
            })

        return basket

    def transaction_customer(self, store):
        # w/c 30 september 2019 uncomment this line...
        # return store.data['customer']

        # ... send delete the following
        return store.data.get('customer', { 'email': None })

    def transaction_extras(self, store):
        return store.data.get('extras', None)

    def transaction_invoice(self, store):
        data = store.data['invoice_address']

        # invoice not requested
        if not data['required']:
            return None

        # invoice requested, populate data
        return {
            'address_1': data['address']['address_1'],
            'address_2': data['address']['address_2'],
            'address_3': data['address']['address_3'],
            'address_4': data['address']['address_4'],
            'address_5': data['address']['address_5'],
            'country_id': data['address']['country']['id'] if isinstance(data['address']['country'], dict) else data['address']['country'],
            'country_state_id': data['address']['country_state']['id'] if isinstance(data['address']['country_state'], dict) else data['address']['country_state'],
            'email': data['address']['email'],
            'phone_1': data['address']['phone_1'],
            'phone_2': data['address']['phone_2'],
            'postcode': data['address']['postcode'],
            'string_1': data['address']['string_1'],
            'string_2': data['address']['string_2'],
            'recipient_name': data['address']['full_name'],
            'recipient_tax_id': data['address']['tax_id']
        }

    def transaction_member(self, store):
        if store.user is None:
            return None

        # retrieve user
        user = get_user_model().objects.get(id=store.user)
        if user.salebox_member_id is None:
            return None

        return {
            "salebox_member_id": str(user.salebox_member_id),
            "total_loyalty": store.data['basket']['loyalty']
        }

    def transaction_meta(self, store):
        checkout_dt = None
        transaction_dt = [
            store.created.year,
            store.created.month,
            store.created.day,
            store.created.hour,
            store.created.minute,
            store.created.second,
            store.created.microsecond
        ]

        if store.payment_received is not None:
            checkout_dt = [
                store.created.year,
                store.created.month,
                store.created.day,
                store.created.hour,
                store.created.minute,
                store.created.second,
                store.created.microsecond
            ]
            transaction_dt = [
                store.payment_received.year,
                store.payment_received.month,
                store.payment_received.day,
                store.payment_received.hour,
                store.payment_received.minute,
                store.payment_received.second,
                store.payment_received.microsecond
            ]

        return {
            'guid': store.visible_id,
            'checkout_dt': checkout_dt,
            'user_id': settings.SALEBOX['MISC']['POS_USER_ID'],
            'utc': transaction_dt
        }

    def transaction_payment(self, store):
        total_price = (
            store.data['basket']['sale_price'] +
            store.data['shipping_method']['price']
        )

        return [
            {
                "amount_change": 0,
                "amount_tendered": total_price,
                "method_id": store.data['payment_method']['id']
            }
        ]

    def transaction_shipping(self, store):
        address = store.data['shipping_address']
        method = store.data['shipping_method']

        # shipping not requested
        if not address['required']:
            return None

        # shipping requested, populate data
        return {
            'address_1': address['address']['address_1'],
            'address_2': address['address']['address_2'],
            'address_3': address['address']['address_3'],
            'address_4': address['address']['address_4'],
            'address_5': address['address']['address_5'],
            'country_id': address['address']['country']['id'] if isinstance(address['address']['country'], dict) else address['address']['country'],
            'country_state_id': address['address']['country_state']['id'] if isinstance(address['address']['country_state'], dict) else address['address']['country_state'],
            'email': address['address']['email'],
            'method_id': method['id'],
            'phone_1': address['address']['phone_1'],
            'phone_2': address['address']['phone_2'],
            'postcode': address['address']['postcode'],
            'price': method['price'],
            'string_1': address['address']['string_1'],
            'string_2': address['address']['string_2'],
            'recipient_name': address['address']['full_name']
        }

    def transaction_total(self, store):
        vat_rate = settings.SALEBOX['MISC']['VAT_RATE']

        # calculate gross + gross_vat_applicable
        gross_total = 0
        gross_vat_applicable = 0
        for b in store.data['basket']['items']:
            gross_total += b['variant']['qty_sale_price']
            if b['variant']['product']['vat_applicable']:
                gross_vat_applicable += b['variant']['qty_sale_price']

        # calculate
        total_vat = round(gross_vat_applicable / (1 + (vat_rate / 100)))
        total_vat = gross_vat_applicable - total_vat
        return {
            'total_gross': gross_total,
            'total_net': gross_total - total_vat,
            'total_vat': total_vat,
            'vat_rate': vat_rate
        }



# Generated by Django 2.1.8 on 2019-10-05 05:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0007_auto_20191005_1228'),
    ]

    operations = [
        migrations.AlterField(
            model_name='countrystate',
            name='code',
            field=models.CharField(blank=True, max_length=5, null=True),
        ),
    ]



import datetime
import requests

from django.conf import settings
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string

try:
    from mailqueue.models import MailerMessage
except:
    pass

from saleboxdjango.models import Event, Member


class SaleboxEventHandler:
    def __init__(self):
        # get list of events
        es = Event \
                .objects \
                .filter(processed_flag=False) \
                .order_by('created')

        # loop through and process them
        for e in es:
            if e.event == 'member_created':
                self.event_member_created(e)
            elif e.event == 'transaction_created':
                self.event_transaction_created(e)
            elif e.event == 'transaction_shipping_packed':
                self.event_transaction_shipping_packed(e)
            elif e.event == 'transaction_shipping_picked':
                self.event_transaction_shipping_picked(e)
            elif e.event == 'transaction_shipping_shipped':
                self.event_transaction_shipping_shipped(e)

    def event_member_created(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_created(self, event):
        try:
            t = self._fetch_transaction(event)
            if t['status'] != 'OK':
                return

            # queue email
            email = self._fetch_transaction_email(t)
            if email is not None:
                self._mailqueue(
                    email,
                    render_to_string(
                        'salebox/email/transaction_created/subject.txt',
                        t['transaction']
                    ),
                    render_to_string(
                        'salebox/email/transaction_created/body.txt',
                        t['transaction']
                    ),
                    render_to_string(
                        'salebox/email/transaction_created/body.html',
                        t['transaction']
                    )
                )

            # mark as processed
            event.processed_flag = True
            event.save()
            self._sync_member_history(event)
        except:
            pass

    def event_transaction_shipping_packed(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_shipping_picked(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_shipping_shipped(self, event):
        try:
            t = self._fetch_transaction(event)
            if t['status'] != 'OK':
                return

            # queue email
            email = self._fetch_transaction_email(t)
            if email is not None:
                self._mailqueue(
                    email,
                    render_to_string(
                        'salebox/email/shipping_shipped/subject.txt',
                        t['transaction']
                    ),
                    render_to_string(
                        'salebox/email/shipping_shipped/body.txt',
                        t['transaction']
                    ),
                    render_to_string(
                        'salebox/email/shipping_shipped/body.html',
                        t['transaction']
                    )
                )

            # mark as processed
            event.processed_flag = True
            event.save()
            self._sync_member_history(event)
        except:
            pass

    def _fetch_member(self, event):
        member = Member \
                    .objects \
                    .filter(salebox_member_id=event.salebox_member_id) \
                    .select_related('parent') \
                    .first()

        user = get_user_model() \
                    .objects \
                    .filter(salebox_member_id=event.salebox_member_id) \
                    .first()

        return {
            'member': member,
            'user': user
        }

    def _fetch_transaction(self, event):
        post = {
            'pos': settings.SALEBOX['API']['KEY'],
            'license': settings.SALEBOX['API']['LICENSE'],
            'platform_type': 'ECOMMERCE',
            'platform_version': '0.1.6',
            'pos_guid': event.transaction_guid
        }

        # do request
        url = '%s/api/pos/v2/transaction/fetch' % settings.SALEBOX['API']['URL']
        try:
            r = requests.post(url, data=post)
            o = r.json()
            o['transaction']['dt'] = datetime.datetime(*o['transaction']['dt'])
            return o
        except:
            return None

    def _fetch_transaction_email(self, transaction):
        if transaction['transaction']['customer_email'] is not None:
            return transaction['transaction']['customer_email']

        if transaction['transaction']['member'] is not None and 'email' in transaction['transaction']['member']:
            return transaction['transaction']['member']['email']

        return None

    def _sync_member_history(self, event):
        if event.salebox_member_id is not None:
            m = Member \
                    .objects \
                    .filter(salebox_member_id=event.salebox_member_id) \
                    .first()
            if m is not None:
                m.transactionhistory_request_sync()

    # optional: for use with django-mail-queue
    def _mailqueue(self, to_address, subject, content, html=None, cc=None, bcc=None, from_address=None):
        try:
            MailerMessage
        except:
            return

        # from address
        if from_address is None:
            from_address = settings.DEFAULT_FROM_EMAIL

        # create message
        msg = MailerMessage()
        msg.from_address = from_address
        msg.subject = subject.replace('\n', '').strip()
        msg.to_address = to_address
        msg.content = content
        msg.app = 'Salebox'

        # optional extras
        if html is not None:
            msg.html_content = html
        if cc is not None:
            msg.cc_address = cc
        if bcc is not None:
            msg.bcc_address = bcc

        # save in queue
        msg.save()



from django.http import JsonResponse
from django.shortcuts import redirect
from django.views.generic import View

from saleboxdjango.lib.address import SaleboxAddress


class SaleboxBaseView(View):
    user_must_be_authenticated = True
    http_method_names = ['post']

    form = None  # django form
    action = None  # name of action performed
    status = None  # action outcome

    redirect = None  # url to redirect to (if applicable)
    state = None  # optional extra string data, passed thru
    results_csv = ''  # csv of requested results (if applicable)
    results = {}  # dict of results to return

    def add_json(self, key):
        if getattr(self, key, None):
            self.results[key] = getattr(self, key)

    def add_querystring(self, key, value):
        if value:
            self.redirect = '%s%s%s=%s' % (
                self.redirect,
                '&' if '?' in self.redirect else '?',
                key,
                value
            )

    def form_valid(self, request):
        self.status = 'success'

    def form_invalid(self, request):
        self.status = 'fail'

    def init_class(self, request):
        pass

    def post(self, request):
        if self.user_must_be_authenticated and not request.user.is_authenticated:
            return JsonResponse({'error': 'unauthenticated'})

        self.form = self.form(request.POST)
        if self.form.is_valid():
            # retrieve form values
            self.redirect = self.form.cleaned_data['redirect']
            self.state = self.form.cleaned_data['state']
            if self.form.cleaned_data['results']:
                self.results_csv = self.form.cleaned_data['results']

            # perform task
            self.init_class(request)
            self.form_valid(request)
        else:
            # handle error
            self.form_invalid(request)

        # redirect if applicable
        if self.redirect:
            self.add_querystring('state', self.state)
            self.add_querystring('action', self.action)
            self.add_querystring('status', self.status)
            return redirect(self.redirect)

        # return json
        self.add_json('state')
        self.add_json('action')
        self.add_json('status')
        return JsonResponse(self.results)



from django.urls import path

# address
from .views.address.add import SaleboxAddressAddView
from .views.address.remove import SaleboxAddressRemoveView
from .views.address.set_default import SaleboxAddressSetDefaultView

# basket
from .views.basket.basket import SaleboxBasketBasketView
from .views.basket.migrate import SaleboxBasketMigrateView
from .views.basket.wishlist import SaleboxBasketWishlistView

# image
from saleboxdjango.views.image.image import salebox_image_view

# rating
from .views.rating.add import SaleboxRatingAddView
from .views.rating.remove import SaleboxRatingRemoveView

# sync
from .views.sync.sync import SaleboxSyncView

urlpatterns = [
    # address
    path('address/add/', SaleboxAddressAddView.as_view()),
    path('address/remove/', SaleboxAddressRemoveView.as_view()),
    path('address/set-default/', SaleboxAddressSetDefaultView.as_view()),

    # basket
    path('basket/basket/', SaleboxBasketBasketView.as_view()),
    path('basket/migrate/', SaleboxBasketMigrateView.as_view()),
    path('basket/wishlist/', SaleboxBasketWishlistView.as_view()),

    # image
    path('img/<slug:imgtype>/<slug:dir>/<int:id>.<slug:hash>.<slug:suffix>', salebox_image_view),

    # rating
    path('rating/add/', SaleboxRatingAddView.as_view()),
    path('rating/remove/', SaleboxRatingRemoveView.as_view()),

    # sync
    path('sync/', SaleboxSyncView),
]



# Generated by Django 2.1.8 on 2019-08-13 06:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='boolean_1',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_2',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_3',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='product',
            name='boolean_4',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='boolean_1',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='boolean_2',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='boolean_3',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='boolean_4',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='boolean_1',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='boolean_2',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='boolean_3',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='boolean_4',
            field=models.BooleanField(default=False),
        ),
    ]



from saleboxdjango.views.basket.base import SaleboxBasketView
from saleboxdjango.forms import SaleboxBasketBasketForm


class SaleboxBasketBasketView(SaleboxBasketView):
    action = 'basket-basket'
    form = SaleboxBasketBasketForm

    def form_valid(self, request):
        try:
            self.sb.update_basket(
                request,
                self.form.cleaned_data['variant_id'],
                self.form.cleaned_data['quantity'],
                self.form.cleaned_data['relative']
            )
            self.status = 'success'

            self.results = self.sb.get_data(
                request,
                self.results_csv,
                self.form.cleaned_data['variant_id']
            )
        except:
            self.status = 'fail'



from django import forms
from django.shortcuts import redirect

from saleboxdjango.forms import SaleboxAddressAddForm
from saleboxdjango.lib.address import SaleboxAddress
from saleboxdjango.models import UserAddress
from saleboxdjango.views.checkout.base import SaleboxCheckoutBaseView


class SaleboxCheckoutShippingAddressForm(forms.Form):
    shipping_address_id = forms.IntegerField()


class SaleboxFormNameForm(forms.Form):
    salebox_form_name = forms.ChoiceField(choices=(
        ('select_address', 'select_address'),
        ('add_shipping', 'add_shipping'),
    ))

class SaleboxCheckoutShippingAddressView(SaleboxCheckoutBaseView):
    initial = {}
    checkout_step = 'shipping_address'
    form_class = SaleboxCheckoutShippingAddressForm
    template_name = 'salebox/checkout/shipping_address.html'

    def get(self, request, *args, **kwargs):
        self.shipping_form = SaleboxAddressAddForm(initial=self.initial)
        return super().get(self, request, *args, **kwargs)

    def get_additional_context_data(self, context):
        sa = SaleboxAddress(self.request.user)

        # shipping form
        context['shipping_form'] = self.shipping_form
        context['shipping_addresses'] = sa.get(
            selected_id=self.sc.data['shipping_address']['id'],
            force_selected=True
        )
        context['shipping_address_extras'] = sa.form_extras(
            country_id=self.shipping_form['country'].value()
        )

        # form control
        context['shipping_address_id'] = self.sc.data['shipping_address']['id']
        return context

    def post(self, request, *args, **kwargs):
        self.shipping_form = SaleboxAddressAddForm(initial=self.initial)

        # which action is being performed?
        form = SaleboxFormNameForm(request.POST)
        if form.is_valid():
            action = form.cleaned_data['salebox_form_name']
        else:
            return self.get(request, *args, **kwargs)

        # action: user has selected an option
        if action == 'select_address':
            form = self.form_class(request.POST)
            if form.is_valid():
                self.sc.set_shipping_address(
                    True,
                    form.cleaned_data['shipping_address_id'],
                    None,
                    request
                )

                r = self.sc.set_completed(self.checkout_step, request)
                if r is None:
                    raise Exception('There is no next checkout step to redirect to...')
                else:
                    return redirect(r)

        # action: user is adding a shipping address
        if action == 'add_shipping':
            self.shipping_form = SaleboxAddressAddForm(
                request.POST,
                initial=self.initial
            )
            if self.shipping_form.is_valid():
                # add address to db
                sa = SaleboxAddress(request.user)
                address = sa.add(self.shipping_form.cleaned_data)
                self.sc.set_shipping_address(True, address.id, None, request)

                # redirect to self to prevent refresh
                return redirect(request.get_full_path())

        # if we've reached this point, some input was invalid...
        # just re-show the page
        return self.form_invalid(form)



from django.http import HttpResponse, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from saleboxdjango.models import CallbackStore, CheckoutStore, CheckoutStoreUpdate


@method_decorator(csrf_exempt, name='dispatch')
class SaleboxCallbackView(View):
    get_redirect = None
    post_redirect = None

    def dispatch(self, request, *args, **kwargs):
        CallbackStore(
            ip_address=request.META['REMOTE_ADDR'],
            method=request.method.lower(),
            post=request.POST
        ).save()

        return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        if self.get_redirect is not None:
            return HttpResponseRedirect(self.get_redirect)
        return HttpResponse('OK')

    def post(self, request):
        if self.post_redirect is not None:
            return HttpResponseRedirect(self.post_redirect)
        return HttpResponse('OK')

    def _get_store(self, uuid, visible_id):
        return CheckoutStore \
                .objects \
                .filter(uuid=uuid) \
                .filter(visible_id=visible_id) \
                .first()

    def _save_store_update(self, store, status, data):
        CheckoutStoreUpdate(
            store=store,
            status=status,
            data=data,
        ).save()



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages

setup(
    name='salebox-django',
    version='0.0.187',
    author=u'Jon Combe',
    author_email='jon@getsalebox.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django-mptt',
        'django-opposable-thumbs',
        'pyShipping-python3',
        'requests'
    ],
    url='https://getsalebox.com',
    license='BSD licence, see LICENCE file',
    description='Django ecommerce powered by Salebox',
    long_description='Connect a standalone Django ecommerce website to a GetSalebox.com backoffice',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    zip_safe=False,
)



from django import forms
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt

from saleboxdjango.models import CheckoutStore
from saleboxdjango.views.callback.callback import SaleboxCallbackView


class SaleboxProvider2C2PCallbackForm(forms.Form):
    order_id = forms.CharField()
    payment_status = forms.CharField()
    user_defined_1 = forms.CharField()


@method_decorator(csrf_exempt, name='dispatch')
class SaleboxProviders2C2PCallbackView(SaleboxCallbackView):
    def post(self, request):
        form = SaleboxProvider2C2PCallbackForm(request.POST)
        if form.is_valid():
            cs = self._get_store(
                form.cleaned_data['user_defined_1'],
                form.cleaned_data['order_id']
            )

            if cs is not None:
                if cs.status < 30:
                    status = form.cleaned_data['payment_status']

                    # success
                    if status == '000':
                        cs.set_status(30)

                    # reject
                    if status in ['002', '003', '999']:
                        cs.set_status(40)

                # save update
                self._save_store_update(cs, cs.status, request.POST)

        # repond
        return super().post(request)



import json

from django.conf import settings
from django.http import JsonResponse
from django.template.loader import render_to_string
from django.utils.crypto import get_random_string
from django.utils.translation import get_language

from saleboxdjango.lib.translation import get_translation, get_translations
from saleboxdjango.models import Country, CountryState, UserAddress


class SaleboxAddress:
    def __init__(self, user, address_group='default'):
        self.user = user
        if self.user.is_authenticated:
            self.query = UserAddress \
                            .objects \
                            .filter(user=user) \
                            .filter(address_group=address_group) \
                            .filter(active_flag=True) \
                            .select_related('country', 'country_state')

    def add(self, values):
        # get country
        if isinstance(values['country'], int):
            values['country'] = \
                Country.objects.get(id=values['country'])

        # get country_state
        if isinstance(values['country_state'], int):
            values['country_state'] = \
                CountryState.objects.get(id=values['country_state'])

        # create address instance
        address = UserAddress(
            default=values['default'] or False,
            address_group=values['address_group'] or 'default',
            full_name=values['full_name'],
            address_1=values['address_1'],
            address_2=values['address_2'],
            address_3=values['address_3'],
            address_4=values['address_4'],
            address_5=values['address_5'],
            country_state=values['country_state'],
            country=values['country'],
            postcode=str(values['postcode'] or '').upper(),
            phone_1=values['phone_1'],
            phone_2=values['phone_2'],
            email=values['email'],
            string_1=values['string_1'],
            string_2=values['string_2'],
            tax_id=values['tax_id']
        )

        # save to db if user is authenticated
        if self.user.is_authenticated:
            address.user=self.user
            address.save()

        return address

    def form_extras(self, country_id=None):
        country_list = self._get_country_list()
        country_state_lookup = self._get_country_state_lookup()

        # build 'current' country_state list
        try:
            current_country_states = country_state_lookup[int(country_id)]
        except:
            current_country_states = []

        # build javascript country_state dict
        js_country_states = \
            json.dumps(country_state_lookup, ensure_ascii=False) \
                .replace('"id": ', 'i:') \
                .replace(', "name": ', ',s:') \
                .replace('}, {', '},{')

        return {
            'country_list': country_list,
            'current_states': current_country_states,
            'js_states': js_country_states
        }

    def get(
            self,
            selected_id=None,
            force_selected=False,
            non_null_fields=[]
        ):
            # return an empty list for non-authenticated users
            if not self.user.is_authenticated:
                return []

            # get addresses
            addresses = self.query.all()
            if 'tax_id' in non_null_fields:
                addresses = addresses.filter(tax_id__isnull=False)

            # add 'selected' value and localised country name
            for a in addresses:
                a.selected = False
                # a.country_name = self._get_country_name(a.country)
                # a.country_state_name = self._get_country_state_name(a.country_state)

            # set selected_id if exists
            has_selected = False
            for a in addresses:
                if a.id == selected_id:
                    a.selected = True
                    has_selected = True

            # force selection of an address
            if force_selected and not has_selected and len(addresses) > 0 :
                # select the default address
                for a in addresses:
                    if a.default:
                        a.selected = True
                        has_selected = True
                        break

                # fall back to the first address
                if not has_selected:
                    addresses[0].selected = True

            # return
            return addresses

    def get_count(self):
        return self.query.all().count()

    def set_default(self, address_id):
        address = self._get_address_by_id(address_id)
        if address is not None and not address.default:
            address.default = True
            address.save()

    def set_inactive(self, address_id):
        address = self._get_address_by_id(address_id)
        if address is not None:
            address.active_flag = False
            address.save()

    def _get_address_by_id(self, address_id):
        return self.query.all().filter(id=address_id).first()

    def _get_allowed_countries(self):
        try:
            return settings.SALEBOX['MISC']['ALLOWED_COUNTRIES']
        except:
            return []

    def _get_country_list(self):
        lang = get_language()

        countries = Country.objects.all()
        allowed_countries = self._get_allowed_countries()
        if len(allowed_countries) > 0:
            countries = countries.filter(id__in=allowed_countries)
        countries = list(countries.values('id', 'code'))

        # get translated name
        for c in countries:
            c['name'] = get_translation(
                lang,
                'place.%s' % c['code'],
                c['code']
            )
            del c['code']

        # sort
        countries = sorted(countries, key=lambda k: k['name'])
        return countries

    def _get_country_state_lookup(self):
        lang = get_language()
        i18n = get_translations()

        states_list = CountryState.objects.all()
        allowed_countries = self._get_allowed_countries()
        if len(allowed_countries) > 0:
            states_list = states_list.filter(country__id__in=allowed_countries)
        states_list = list(states_list.values('id', 'country__id', 'full_code'))

        # create a lookup
        lookup = {}
        for s in states_list:
            lookup[s['id']] = {
                'country': s['country__id'],
                'name': get_translation(
                    lang,
                    'place.%s' % s['full_code'],
                    translations=i18n
                )
            }

        # create dict of states per country
        countries = {}
        for key in lookup:
            if lookup[key]['country'] not in countries:
                countries[lookup[key]['country']] = []
            countries[lookup[key]['country']].append({
                'id': key,
                'name': lookup[key]['name']
            })

        # sort each country
        if not lang.startswith('en'):
            for c in countries:
                countries[c] = sorted(countries[c], key=lambda k: k['name'])

        return countries



from django import template
from django.utils.translation import get_language

from saleboxdjango.lib.translation import get_translation
from saleboxdjango.models import Country, CountryState

register = template.Library()

@register.simple_tag
def sb_country_name(country, default='', lang=None):
    if isinstance(country, dict):
        country = country['id']
    if isinstance(country, int):
        country = Country.objects.get(id=country)
    if not isinstance(country, Country):
        return default

    if lang is None:
        lang = (get_language()).lower().split('-')[0]

    return get_translation(
        lang,
        'place.%s' % country.code,
        default
    )


@register.simple_tag
def sb_country_state_name(country_state, default='', lang=None):
    if isinstance(country_state, dict):
        country_state = country_state['id']
    if isinstance(country_state, int):
        country_state = CountryState.objects.get(id=country_state)
    if not isinstance(country_state, CountryState):
        return default

    if lang is None:
        lang = (get_language()).lower().split('-')[0]

    return get_translation(
        lang,
        'place.%s' % country_state.full_code,
        default
    )



# Generated by Django 2.1.8 on 2019-10-05 08:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0008_auto_20191005_1229'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='countrystatetranslation',
            name='state',
        ),
        migrations.RemoveField(
            model_name='countrytranslation',
            name='country',
        ),
        migrations.DeleteModel(
            name='CountryStateTranslation',
        ),
        migrations.DeleteModel(
            name='CountryTranslation',
        ),
    ]



# Generated by Django 2.1.8 on 2019-06-15 01:36

from django.conf import settings
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import mptt.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=20)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'Product Attribute',
                'ordering': ['code'],
            },
        ),
        migrations.CreateModel(
            name='AttributeItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=100)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('delete_dt', models.DateTimeField(blank=True, null=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('attribute', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Attribute')),
            ],
            options={
                'verbose_name': 'Product Attribute Item',
                'ordering': ['value'],
            },
        ),
        migrations.CreateModel(
            name='BasketWishlist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('session', models.CharField(blank=True, max_length=32, null=True)),
                ('basket_flag', models.BooleanField(default=True)),
                ('quantity', models.IntegerField(default=1)),
                ('weight', models.IntegerField(default=0)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='CallbackStore',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ip_address', models.GenericIPAddressField()),
                ('method', models.CharField(max_length=7)),
                ('post', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='CheckoutStore',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(db_index=True)),
                ('visible_id', models.CharField(max_length=14, unique=True)),
                ('user', models.IntegerField(blank=True, null=True)),
                ('gateway_code', models.CharField(max_length=12)),
                ('status', models.IntegerField(choices=[(10, 'New: Pending send to gateway'), (20, 'Pending: Awaiting gateway response'), (30, 'Success: Pending POST to Salebox'), (31, 'Success: Successfully POSTed to Salebox'), (40, 'Rejected: Gateway rejected payment'), (50, 'Timeout: Gateway did not respond in an acceptable time period')])),
                ('data', django.contrib.postgres.fields.jsonb.JSONField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='CheckoutStoreUpdate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(10, 'New: Pending send to gateway'), (20, 'Pending: Awaiting gateway response'), (30, 'Success: Pending POST to Salebox'), (31, 'Success: Successfully POSTed to Salebox'), (40, 'Rejected: Gateway rejected payment'), (50, 'Timeout: Gateway did not respond in an acceptable time period')])),
                ('data', django.contrib.postgres.fields.jsonb.JSONField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('store', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.CheckoutStore')),
            ],
        ),
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code_2', models.CharField(blank=True, max_length=2, null=True)),
                ('code_3', models.CharField(blank=True, max_length=3, null=True)),
                ('name', models.CharField(blank=True, max_length=50, null=True)),
                ('default', models.BooleanField(default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name_plural': 'Countries',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='CountryState',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code_2', models.CharField(blank=True, max_length=2, null=True)),
                ('name', models.CharField(max_length=50)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('country', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Country')),
            ],
            options={
                'verbose_name_plural': 'Country States',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='CountryStateTranslation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language', models.CharField(max_length=7)),
                ('value', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('state', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.CountryState')),
            ],
            options={
                'verbose_name': 'Country State Translations',
                'ordering': ['language', 'state'],
            },
        ),
        migrations.CreateModel(
            name='CountryTranslation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language', models.CharField(max_length=7)),
                ('value', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(auto_now=True)),
                ('country', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Country')),
            ],
            options={
                'verbose_name': 'Country Translations',
                'ordering': ['language', 'country'],
            },
        ),
        migrations.CreateModel(
            name='DiscountGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=25)),
                ('group_type', models.CharField(choices=[('S', 'Seasonal'), ('M', 'Manual')], default='M', max_length=1)),
                ('operational_flag', models.BooleanField(default=False)),
                ('active_flag', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'Discount Group',
            },
        ),
        migrations.CreateModel(
            name='DiscountRuleset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=25)),
                ('discount_type', models.CharField(choices=[('flat_percent', 'Flat Percentage')], default='flat_percent', max_length=12)),
                ('value', models.IntegerField(blank=True, null=True)),
                ('start_dt', models.DateTimeField(blank=True, null=True)),
                ('end_dt', models.DateTimeField(blank=True, null=True)),
                ('operational_flag', models.BooleanField(default=True)),
                ('active_flag', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.DiscountGroup')),
            ],
            options={
                'verbose_name': 'Discount Ruleset',
            },
        ),
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event', models.CharField(max_length=50)),
                ('transaction_guid', models.CharField(blank=True, max_length=20, null=True)),
                ('salebox_member_id', models.UUIDField(blank=True, default=uuid.uuid4, null=True)),
                ('processed_flag', models.BooleanField(default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='LastUpdate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=36)),
                ('value', models.FloatField(default=0.0)),
            ],
            options={
                'verbose_name': 'Last Update',
                'ordering': ['code'],
            },
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('guid', models.CharField(db_index=True, max_length=25)),
                ('salebox_member_id', models.UUIDField(blank=True, db_index=True, null=True)),
                ('gender', models.CharField(blank=True, choices=[('M', 'Male'), ('F', 'Female'), ('U', 'Unspecified')], max_length=1, null=True)),
                ('title', models.IntegerField(blank=True, choices=[(1, 'Mr'), (2, 'Mrs'), (3, 'Miss')], null=True)),
                ('name_first', models.CharField(blank=True, max_length=20, null=True)),
                ('name_last', models.CharField(blank=True, max_length=30, null=True)),
                ('date_of_birth', models.DateField(blank=True, null=True)),
                ('address_1', models.CharField(blank=True, max_length=255, null=True)),
                ('address_2', models.CharField(blank=True, max_length=255, null=True)),
                ('address_3', models.CharField(blank=True, max_length=255, null=True)),
                ('address_4', models.CharField(blank=True, max_length=255, null=True)),
                ('address_5', models.CharField(blank=True, max_length=255, null=True)),
                ('postcode', models.CharField(blank=True, max_length=12, null=True)),
                ('phone_1', models.CharField(blank=True, max_length=20, null=True)),
                ('phone_2', models.CharField(blank=True, max_length=20, null=True)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('id_card', models.CharField(blank=True, max_length=20, null=True)),
                ('status', models.CharField(choices=[('A', 'Active'), ('R', 'Resigned'), ('S', 'Suspended'), ('T', 'Terminated')], default='A', max_length=1)),
                ('active_flag', models.BooleanField(default=True)),
                ('join_date', models.DateField(blank=True, null=True)),
                ('string_1', models.CharField(blank=True, max_length=255, null=True)),
                ('string_2', models.CharField(blank=True, max_length=255, null=True)),
                ('string_3', models.CharField(blank=True, max_length=255, null=True)),
                ('string_4', models.CharField(blank=True, max_length=255, null=True)),
                ('string_5', models.CharField(blank=True, max_length=255, null=True)),
                ('string_6', models.CharField(blank=True, max_length=255, null=True)),
                ('string_7', models.CharField(blank=True, max_length=255, null=True)),
                ('string_8', models.CharField(blank=True, max_length=255, null=True)),
                ('string_9', models.CharField(blank=True, max_length=255, null=True)),
                ('string_10', models.CharField(blank=True, max_length=255, null=True)),
                ('string_11', models.CharField(blank=True, max_length=255, null=True)),
                ('string_12', models.CharField(blank=True, max_length=255, null=True)),
                ('string_13', models.CharField(blank=True, max_length=255, null=True)),
                ('string_14', models.CharField(blank=True, max_length=255, null=True)),
                ('boolean_1', models.BooleanField(default=False)),
                ('boolean_2', models.BooleanField(default=False)),
                ('boolean_3', models.BooleanField(default=False)),
                ('boolean_4', models.BooleanField(default=False)),
                ('boolean_5', models.BooleanField(default=False)),
                ('boolean_6', models.BooleanField(default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('country', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Country')),
                ('country_state', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.CountryState')),
            ],
            options={
                'verbose_name': 'Member',
                'ordering': ['guid'],
            },
        ),
        migrations.CreateModel(
            name='MemberGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('flat_discount_percentage', models.FloatField(default=0)),
                ('can_be_parent', models.BooleanField(default=True)),
                ('default_group', models.BooleanField(default=False)),
                ('active_flag', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'Member Group',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150)),
                ('string_1', models.CharField(blank=True, max_length=150, null=True)),
                ('string_2', models.CharField(blank=True, max_length=150, null=True)),
                ('string_3', models.CharField(blank=True, max_length=150, null=True)),
                ('string_4', models.CharField(blank=True, max_length=150, null=True)),
                ('sold_by', models.CharField(choices=[('item', 'item'), ('weight', 'weight')], default='item', max_length=6)),
                ('vat_applicable', models.BooleanField(default=True)),
                ('image', models.CharField(blank=True, max_length=70, null=True)),
                ('local_image', models.CharField(blank=True, max_length=25, null=True)),
                ('inventory_flag', models.BooleanField(default=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=100, null=True)),
                ('bestseller_rank', models.IntegerField(default=0)),
                ('active_flag', models.BooleanField(default=True)),
                ('i18n', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('rating_score', models.IntegerField(default=0)),
                ('rating_vote_count', models.IntegerField(default=0)),
                ('attribute_1', models.ManyToManyField(blank=True, related_name='product_attr_1', to='saleboxdjango.AttributeItem')),
                ('attribute_10', models.ManyToManyField(blank=True, related_name='product_attr_10', to='saleboxdjango.AttributeItem')),
                ('attribute_2', models.ManyToManyField(blank=True, related_name='product_attr_2', to='saleboxdjango.AttributeItem')),
                ('attribute_3', models.ManyToManyField(blank=True, related_name='product_attr_3', to='saleboxdjango.AttributeItem')),
                ('attribute_4', models.ManyToManyField(blank=True, related_name='product_attr_4', to='saleboxdjango.AttributeItem')),
                ('attribute_5', models.ManyToManyField(blank=True, related_name='product_attr_5', to='saleboxdjango.AttributeItem')),
                ('attribute_6', models.ManyToManyField(blank=True, related_name='product_attr_6', to='saleboxdjango.AttributeItem')),
                ('attribute_7', models.ManyToManyField(blank=True, related_name='product_attr_7', to='saleboxdjango.AttributeItem')),
                ('attribute_8', models.ManyToManyField(blank=True, related_name='product_attr_8', to='saleboxdjango.AttributeItem')),
                ('attribute_9', models.ManyToManyField(blank=True, related_name='product_attr_9', to='saleboxdjango.AttributeItem')),
            ],
            options={
                'verbose_name': 'Product',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ProductCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('short_name', models.CharField(max_length=30)),
                ('name', models.CharField(max_length=100)),
                ('image', models.CharField(blank=True, max_length=70, null=True)),
                ('local_image', models.CharField(blank=True, max_length=25, null=True)),
                ('seasonal_flag', models.BooleanField(default=False)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=100, null=True)),
                ('slug_path', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('active_flag', models.BooleanField(default=True)),
                ('i18n', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('lft', models.PositiveIntegerField(db_index=True, editable=False)),
                ('rght', models.PositiveIntegerField(db_index=True, editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(db_index=True, editable=False)),
                ('parent', mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='saleboxdjango.ProductCategory')),
            ],
            options={
                'verbose_name': 'Product Category',
                'verbose_name_plural': 'Product Categories',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ProductVariant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, default='', max_length=150)),
                ('bo_name', models.CharField(blank=True, default='', max_length=200)),
                ('plu', models.CharField(blank=True, default='', max_length=25)),
                ('sku', models.CharField(blank=True, default='', max_length=25)),
                ('color', models.CharField(blank=True, default='', max_length=50)),
                ('size', models.CharField(blank=True, default='', max_length=20)),
                ('size_order', models.FloatField(default=0)),
                ('size_uom', models.CharField(blank=True, choices=[('', 'n/a'), ('g', 'g'), ('kg', 'kg'), ('ml', 'ml')], default='', max_length=2)),
                ('price', models.IntegerField(null=True)),
                ('sale_price', models.IntegerField(default=0)),
                ('sale_percent', models.IntegerField(default=0)),
                ('barcode', models.CharField(blank=True, default='', max_length=50)),
                ('available_to_order', models.BooleanField(default=True)),
                ('available_on_pos', models.BooleanField(default=True)),
                ('available_on_ecom', models.BooleanField(default=True)),
                ('shelf_expiry_type', models.CharField(default='manual', max_length=12)),
                ('shelf_life_days', models.IntegerField(blank=True, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=150, null=True)),
                ('image', models.CharField(blank=True, max_length=70, null=True)),
                ('local_image', models.CharField(blank=True, max_length=20, null=True)),
                ('unique_string', models.CharField(blank=True, max_length=255)),
                ('shipping_weight', models.IntegerField(blank=True, null=True)),
                ('loyalty_points', models.FloatField(blank=True, null=True)),
                ('member_discount_applicable', models.BooleanField(default=True)),
                ('string_1', models.CharField(blank=True, max_length=150, null=True)),
                ('string_2', models.CharField(blank=True, max_length=150, null=True)),
                ('string_3', models.CharField(blank=True, max_length=150, null=True)),
                ('string_4', models.CharField(blank=True, max_length=150, null=True)),
                ('warehouse_location', models.CharField(blank=True, max_length=50, null=True)),
                ('int_1', models.IntegerField(blank=True, null=True)),
                ('int_2', models.IntegerField(blank=True, null=True)),
                ('int_3', models.IntegerField(blank=True, null=True)),
                ('int_4', models.IntegerField(blank=True, null=True)),
                ('date_1', models.DateField(blank=True, null=True)),
                ('date_2', models.DateField(blank=True, null=True)),
                ('active_flag', models.BooleanField(default=True)),
                ('ecommerce_description', models.TextField(blank=True, null=True)),
                ('ecommerce_low_stock_threshold', models.IntegerField(default=10)),
                ('bestseller_rank', models.IntegerField(default=0)),
                ('default_image', models.CharField(blank=True, max_length=35, null=True)),
                ('i18n', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('stock_count', models.IntegerField(default=0)),
                ('stock_checked_out', models.IntegerField(default=0)),
                ('stock_total', models.IntegerField(default=0)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('rating_score', models.IntegerField(default=0)),
                ('rating_vote_count', models.IntegerField(default=0)),
                ('name_sorted', models.IntegerField(db_index=True, default=0)),
                ('attribute_1', models.ManyToManyField(blank=True, related_name='variant_attr_1', to='saleboxdjango.AttributeItem')),
                ('attribute_10', models.ManyToManyField(blank=True, related_name='variant_attr_10', to='saleboxdjango.AttributeItem')),
                ('attribute_2', models.ManyToManyField(blank=True, related_name='variant_attr_2', to='saleboxdjango.AttributeItem')),
                ('attribute_3', models.ManyToManyField(blank=True, related_name='variant_attr_3', to='saleboxdjango.AttributeItem')),
                ('attribute_4', models.ManyToManyField(blank=True, related_name='variant_attr_4', to='saleboxdjango.AttributeItem')),
                ('attribute_5', models.ManyToManyField(blank=True, related_name='variant_attr_5', to='saleboxdjango.AttributeItem')),
                ('attribute_6', models.ManyToManyField(blank=True, related_name='variant_attr_6', to='saleboxdjango.AttributeItem')),
                ('attribute_7', models.ManyToManyField(blank=True, related_name='variant_attr_7', to='saleboxdjango.AttributeItem')),
                ('attribute_8', models.ManyToManyField(blank=True, related_name='variant_attr_8', to='saleboxdjango.AttributeItem')),
                ('attribute_9', models.ManyToManyField(blank=True, related_name='variant_attr_9', to='saleboxdjango.AttributeItem')),
                ('product', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Product')),
            ],
            options={
                'verbose_name': 'Product Variant',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='ProductVariantImage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('img', models.CharField(default='', max_length=100)),
                ('local_img', models.CharField(blank=True, max_length=25, null=True)),
                ('img_height', models.IntegerField(default=0)),
                ('img_width', models.IntegerField(default=0)),
                ('title', models.CharField(blank=True, max_length=150, null=True)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
                ('order', models.IntegerField(blank=True, null=True)),
                ('active_flag', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('variant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.ProductVariant')),
            ],
            options={
                'ordering': ('order',),
            },
        ),
        migrations.CreateModel(
            name='ProductVariantRating',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('rating', models.IntegerField(default=50)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                ('variant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.ProductVariant')),
            ],
            options={
                'verbose_name': 'Product Variant Rating',
            },
        ),
        migrations.CreateModel(
            name='UserAddress',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('default', models.BooleanField(default=False)),
                ('address_group', models.CharField(default='default', max_length=10)),
                ('full_name', models.CharField(max_length=150)),
                ('address_1', models.CharField(max_length=150)),
                ('address_2', models.CharField(blank=True, max_length=150, null=True)),
                ('address_3', models.CharField(blank=True, max_length=150, null=True)),
                ('address_4', models.CharField(blank=True, max_length=150, null=True)),
                ('address_5', models.CharField(blank=True, max_length=150, null=True)),
                ('postcode', models.CharField(blank=True, max_length=12, null=True)),
                ('phone_1', models.CharField(blank=True, max_length=20, null=True)),
                ('phone_2', models.CharField(blank=True, max_length=20, null=True)),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('string_1', models.CharField(blank=True, max_length=255, null=True)),
                ('string_2', models.CharField(blank=True, max_length=255, null=True)),
                ('tax_id', models.CharField(blank=True, max_length=36, null=True)),
                ('active_flag', models.BooleanField(default=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_update', models.DateTimeField(auto_now=True)),
                ('country', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Country')),
                ('country_state', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.CountryState')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'User Address',
                'ordering': ['-default', 'full_name', 'address_1'],
            },
        ),
        migrations.AddField(
            model_name='product',
            name='category',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.ProductCategory'),
        ),
        migrations.AddField(
            model_name='member',
            name='group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.MemberGroup'),
        ),
        migrations.AddField(
            model_name='member',
            name='group_when_created',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='group_when_created', to='saleboxdjango.MemberGroup'),
        ),
        migrations.AddField(
            model_name='member',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.Member'),
        ),
        migrations.AddField(
            model_name='discountruleset',
            name='product_variant',
            field=models.ManyToManyField(blank=True, to='saleboxdjango.ProductVariant'),
        ),
        migrations.AddField(
            model_name='basketwishlist',
            name='variant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='saleboxdjango.ProductVariant'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['bestseller_rank', 'name_sorted'], name='saleboxdjan_bestsel_5ad3e6_idx'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['-bestseller_rank', 'name_sorted'], name='saleboxdjan_bestsel_e36854_idx'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['sale_price', 'name_sorted'], name='saleboxdjan_sale_pr_1681f7_idx'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['-sale_price', 'name_sorted'], name='saleboxdjan_sale_pr_90497c_idx'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['rating_score', 'name_sorted'], name='saleboxdjan_rating__c1dad4_idx'),
        ),
        migrations.AddIndex(
            model_name='productvariant',
            index=models.Index(fields=['-rating_score', 'name_sorted'], name='saleboxdjan_rating__4d1500_idx'),
        ),
    ]



from saleboxdjango.lib.binpack import BinPack


class SaleboxShippingOptions:
    DEFAULT_SELECT = 'FIRST'  # or 'CHEAPEST' or None
    REMOVE_UNAVAILABLE = True
    SORT_BY = 'PRICE_ASC'  # 'LABEL' or None

    # variant variable names
    SHIPPING_WIDTH = 'int_1'
    SHIPPING_HEIGHT = 'int_2'
    SHIPPING_DEPTH = 'int_3'
    SHIPPING_WEIGHT = 'shipping_weight'

    def get_options(self):
        # This is the function you need to replace.
        # Add your own functions here, e.g.
        #
        #   return [
        #     self._example_option_1(),
        #     self._example_option_2(),
        #   ]
        #
        # Useful vars:
        #   self.checkout['shipping_address']['address']['country']
        #   self.checkout['shipping_address']['address']['country_state']
        #
        return [
            self._example_option_1(),
            self._example_option_2(),
        ]

    def go(self, request, checkout, context):
        self.request = request
        self.checkout = checkout
        self.context = context

        # get packages into a handy list for bin packing
        self.meta = {
            'total_items': 0,
            'total_items_onshelf': 0,
            'total_items_preorder': 0,
            'total_weight': 0,
            'total_weight_onshelf': 0,
            'total_weight_preorder': 0,
            'items': [],
            'items_onshelf': [],
            'items_preorder': []
        }
        for b in checkout['basket']['items']:
            for i in range(0, b['qty']):
                self.meta['total_items'] += 1
                self.meta['total_weight'] += b['variant'][self.SHIPPING_WEIGHT] or 0
                self.meta['items'].append({
                    'variant_id': b['variant']['id'],
                    'width': b['variant'][self.SHIPPING_WIDTH],
                    'height': b['variant'][self.SHIPPING_HEIGHT],
                    'depth': b['variant'][self.SHIPPING_DEPTH],
                    'weight': b['variant'][self.SHIPPING_WEIGHT],
                })

                if b['variant']['preorder_flag']:
                    self.meta['total_items_preorder'] += 1
                    self.meta['total_weight_preorder'] += b['variant'][self.SHIPPING_WEIGHT] or 0
                    self.meta['items_preorder'].append({
                        'variant_id': b['variant']['id'],
                        'width': b['variant'][self.SHIPPING_WIDTH],
                        'height': b['variant'][self.SHIPPING_HEIGHT],
                        'depth': b['variant'][self.SHIPPING_DEPTH],
                        'weight': b['variant'][self.SHIPPING_WEIGHT],
                    })
                else:
                    self.meta['total_items_onshelf'] += 1
                    self.meta['total_weight_onshelf'] += b['variant'][self.SHIPPING_WEIGHT] or 0
                    self.meta['items_onshelf'].append({
                        'variant_id': b['variant']['id'],
                        'width': b['variant'][self.SHIPPING_WIDTH],
                        'height': b['variant'][self.SHIPPING_HEIGHT],
                        'depth': b['variant'][self.SHIPPING_DEPTH],
                        'weight': b['variant'][self.SHIPPING_WEIGHT],
                    })

        # build options
        opts = self.get_options()

        # remove nulls
        # if an shipping option is not available, simply return None. However...
        opts = [o for o in opts if o is not None]

        # optional: remove unavailable
        # you *may* want to keep unavailable options in the list but greyed out,
        # in which case, return the dict with ['available'] = False
        if self.REMOVE_UNAVAILABLE:
            opts = [o for o in opts if o['available'] == True]

        # set selected value (if exists)
        if self.checkout['shipping_method']['id'] is not None:
            option = next(o for o in opts if o['id'] == self.checkout['shipping_method']['id'])
            option['selected'] = True

        # optional: sorting
        if self.SORT_BY == 'PRICE_ASC':
            opts = sorted(opts, key=lambda k: k['price'])
        elif self.SORT_BY == 'LABEL':
            opts = sorted(opts, key=lambda k: k['label']['label'])

        # optional: default select
        if len(opts) > 0:
            if self.DEFAULT_SELECT == 'CHEAPEST':
                tmp = sorted(opts, key=lambda k: k['price'])
                id = tmp[0]['id']
                option = next(o for o in opts if o['id'] == id)
                option['selected'] = True
            elif self.DEFAULT_SELECT == 'FIRST':
                opts[0]['selected'] = True

        # calculate combined_price
        for o in opts:
            o['combined_price'] = (self.checkout['basket']['sale_price'] + o['price'])

        # return
        self.context['shipping_options'] = opts
        return self.context

    def init_option(
            self,
            id,
            code,
            label,
            remarks,
            service,
            meta=None
        ):
        return {
            'available': True,
            'code': code,
            'id': id,
            'label': {
                'label': label,  # e.g. 'Post Office'
                'remarks': remarks,  # e.g. '2-3 days'
                'service': service  # 'ExpressPost'
            },
            'meta': meta,
            'price': 0,
            'selected': False
        }

    def _do_binpack(self, containers, packages):
        # containers example:
        # [
        #   ['name', price, width, height, depth],
        #   ['name', price, width, height, depth],
        # ]
        bp = BinPack()

        # add containers
        for c in containers:
            bp.add_bin(c[0], c[1], c[2], c[3], c[4])

        # add packages
        for p in packages:
            bp.add_package(
                p['variant_id'],
                p['width'] or 0,
                p['height'] or 0,
                p['depth'] or 0,
            )

        return bp.go()

    def _example_option_1(self):
        method = self.init_option(
            1,
            'postoffice',
            'Post Office',
            '2-4 days',
            'Surface mail'
        )

        method['price'] = 10000
        return method

    def _example_option_2(self):
        method = self.init_option(
            2,
            'postofficeexpress',
            'Post Office',
            '1 day',
            'NextDay'
        )

        method['price'] = 25000
        return method



from django.shortcuts import redirect
from django.views.generic import TemplateView

from saleboxdjango.lib.basket import SaleboxBasket


class SaleboxAccountWishlistView(TemplateView):
    template_name = 'salebox/account/wishlist.html'

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        sb = SaleboxBasket(request)
        context['data'] = sb.get_raw_data()

        return self.render_to_response(context)



from functools import reduce
import math
import operator
import re

from django.conf import settings
from django.core.cache import cache
from django.db import connection
from django.db.models import Case, F, Q, Value, When
from django.http import Http404

from saleboxdjango.lib.common import fetchsinglevalue, \
    dictfetchall, get_rating_display

from saleboxdjango.models import Attribute, AttributeItem, \
    Product, ProductCategory, ProductVariant, ProductVariantRating


class SaleboxProduct:
    def __init__(self, active_status='active_only'):
        # product filters
        self.query = ProductVariant.objects
        self.active_status = active_status
        self.excludes = {}
        self.exclude_product_ids = []
        self.exclude_productvariant_ids = []
        self.filters = {}
        self.min_price = None
        self.max_price = None
        self.max_result_count = None
        self.order = []
        self.order_related = None
        self.prefetch_product_attributes = []
        self.prefetch_variant_attributes = []

        # pagination
        self.offset = 0
        self.page_number = 1
        self.limit = 50
        self.items_per_page = 50
        self.pagination_url_prefix = ''

        # misc
        self.fetch_user_ratings = True
        self.flat_discount = 0
        self.flat_member_discount = 0

    def get_list(self, request=None):
        # TODO: retrieve from cache
        #
        #
        data = None

        # cache doesn't exist, build it...
        if data is None:
            # retrieve list of variant IDs (one variant per product)
            # which match our criteria
            self.query = \
                self.query \
                    .exclude(product__id__in=self.exclude_product_ids) \
                    .exclude(**self.excludes) \
                    .filter(**self.filters) \
                    .order_by('product__id', 'price') \
                    .distinct('product__id') \
                    .values_list('id', flat=True)

            variant_ids = self._retrieve_variant_ids(do_exclude=True)

            if self.order_related is not None:
                variant_ids = self._sort_by_related_items(variant_ids)

            data = {
                'variant_ids': variant_ids,
                'qs': self._retrieve_results(variant_ids, preserve_order=self.order_related is not None)
            }

            # TODO: save data to cache
            #
            #

        # pagination calculations
        if self.max_result_count:
            total_results = min(len(data['variant_ids']), self.max_result_count)
        else:
            total_results = len(data['variant_ids'])
        number_of_pages = math.ceil(total_results / self.items_per_page)

        if request is None:
            products = data['qs']
        else:
            products = self.retrieve_user_interaction(request, data['qs'])

        # create output dict
        return {
            'count': {
                'from': self.offset + 1,
                'to': self.offset + len(data['qs']),
                #'total': len(data['variant_ids']),
                'total': total_results
            },
            'pagination': {
                'page_number': self.page_number,
                'number_of_pages': number_of_pages,
                'page_range': range(1, number_of_pages + 1),
                'has_previous': self.page_number > 1,
                'previous': self.page_number - 1,
                'has_next': self.page_number < number_of_pages,
                'next': self.page_number + 1,
                'url_prefix': self.pagination_url_prefix,
                'valid': self.page_number <= max(number_of_pages, 1)
            },
            'products': products
        }

    def get_single(self, request, id, slug):
        self.query = \
            self.query \
                .filter(id=id) \
                .filter(slug=slug) \
                .values_list('id', flat=True)

        # ensure variant exists
        variant_ids = self._retrieve_variant_ids(do_exclude=False)
        if len(variant_ids) == 0:
            raise Http404

        # retrieve variant
        return self.retrieve_user_interaction(
            request,
            self._retrieve_results(variant_ids)
        )[0]

    def retrieve_user_interaction(self, request, variants):
        # get user ratings
        rating_dict = {}
        if self.fetch_user_ratings and request.user.is_authenticated:
            ratings = ProductVariantRating \
                        .objects \
                        .filter(variant__id__in=[pv.id for pv in variants]) \
                        .filter(user=request.user)
            for r in ratings:
                rating_dict[r.variant.id] = r.rating

        # get basket / wishlist flags
        for pv in variants:
            if str(pv.id) in request.session['saleboxbasket']['basket']['lookup']:
                pv.basket_qty = request.session['saleboxbasket']['basket']['lookup'][str(pv.id)]['qty']
            else:
                pv.basket_qty = 0

            pv.in_wishlist = str(pv.id) in request.session['saleboxbasket']['wishlist']['order']

            if pv.id in rating_dict:
                pv.user_rating = get_rating_display(rating_dict[pv.id], 1)
            else:
                pv.user_rating = None

        return variants

    def set_exclude_product_ids(self, id_list):
        if isinstance(id_list, int):
            id_list = [id_list]
        self.exclude_product_ids += id_list

    def set_exclude_productvariant_ids(self, id_list):
        if isinstance(id_list, int):
            id_list = [id_list]
        self.exclude_productvariant_ids += id_list

    def set_minimum_stock_total(self, minimum_stock, allow_preorder=True):
        if allow_preorder:
            self.query = self.query.filter(
                Q(stock_total__gte=minimum_stock) |
                Q(preorder_flag=True) |
                Q(product__inventory_flag=False)
            )
        else:
            self.query = self.query.filter(
                Q(stock_total__gte=minimum_stock) |
                Q(product__inventory_flag=False)
            )

    def set_prefetch_product_attributes(self, numbers):
        if isinstance(numbers, int):
            numbers = [numbers]
        self.prefetch_product_attributes = [
            'product__attribute_%s' % i for i in numbers
        ]

    def set_prefetch_variant_attributes(self, numbers):
        if isinstance(numbers, int):
            numbers = [numbers]
        self.prefetch_variant_attributes = [
            'attribute_%s' % i for i in numbers
        ]

    def set_active_status(self):
        # I can think of no reason for this to ever be set to anything
        # other than 'active_only' but include this here so it doesn't
        # bite us later
        if self.active_status == 'active_only':
            self.query = \
                self.query \
                    .filter(active_flag=True) \
                    .filter(available_on_ecom=True) \
                    .filter(product__active_flag=True) \
                    .filter(product__category__active_flag=True)

        elif self.active_status == 'all':
            pass

    def set_category(self, category, include_child_categories=True):
        if include_child_categories:
            id_list = category \
                        .get_descendants(include_self=True) \
                        .values_list('id', flat=True)
        else:
            if isinstance(category, int):
                id_list = [category]
            else:
                id_list = [category.id]

        self.set_categories(id_list)

    def set_categories(self, category_list):
        self.query = self.query.filter(product__category__in=category_list)

    def set_discount_only(self):
        self.query = self.query.filter(sale_price__lt=F('price'))

    def set_fetch_user_ratings(self, value):
        self.fetch_user_ratings = value

    def set_flat_discount(self, percent):
        self.flat_discount = percent

    def set_flat_member_discount(self, percent):
        self.flat_member_discount = percent

    def set_max_price(self, maximun):
        self.query = self.query.filter(sale_price__lte=maximun)

    def set_max_result_count(self, i):
        self.max_result_count = i

    def set_max_sale_percent(self, maximum):
        self.query = self.query.filter(sale_percent__lte=maximun)

    def set_min_price(self, minimun):
        self.query = self.query.filter(sale_price__gte=minimum)

    def set_min_sale_percent(self, minimum):
        self.query = self.query.filter(sale_percent__gte=minimum)

    def set_order_custom(self, order):
        self.order = list(order)
        if isinstance(self.order, str):
            self.order = [self.order]

    def set_order_preset(self, preset):
        self.order = {
            'bestseller_low_to_high': ['bestseller_rank', 'name_sorted'],
            'bestseller_high_to_low': ['-bestseller_rank', 'name_sorted'],
            'name_low_to_high': ['name_sorted'],
            'name_high_to_low': ['-name_sorted'],
            'price_low_to_high': ['sale_price', 'name_sorted'],
            'price_high_to_low': ['-sale_price', 'name_sorted'],
            'rating_low_to_high': ['rating_score', 'rating_vote_count', 'name_sorted'],
            'rating_high_to_low': ['-rating_score', '-rating_vote_count', 'name_sorted'],
        }[preset]

    def set_order_related(self, root_category, fields, string=None, string_weight=10):
        self.order_related = {
            'category_ids': list(root_category.get_descendants(include_self=True).values_list('id', flat=True)),
            'fields': fields,
            'string': {
                'string': string,
                'weight': string_weight
            }
        }

    def set_order_related_from_variant(self, variant, fields):
        # exclude this variant
        self.set_exclude_productvariant_ids([variant.id])

        # generate sequence
        for f in fields:
            if f['type'] == 'category':
                f['values'] = [variant.product.category.id]
            if f['type'] == 'product':
                f['values'] = list(getattr(variant.product, 'attribute_%s' % f['id']).all().values_list('id', flat=True))
            if f['type'] == 'variant':
                f['values'] = list(getattr(variant, 'attribute_%s' % f['id']).all().values_list('id', flat=True))

        # remove "empties"
        fields = [f for f in fields if len(f['values']) > 0]

        # pass to main function
        self.set_order_related(
            variant.product.category.get_root(),
            fields,
            variant.name
        )

    def set_pagination(self, page_number, items_per_page, url_prefix):
        self.page_number = page_number
        self.offset = (page_number - 1) * items_per_page
        self.limit = self.offset + items_per_page
        self.items_per_page = items_per_page
        self.pagination_url_prefix = url_prefix

    def set_product_attribute_include(self, attribute_number, value):
        key = 'product__attribute_%s' % attribute_number
        self.query = self.query.filter(**{key: value})

    def set_product_attribute_include_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'product__attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.query = self.query.filter(**{key: field_value})

    def set_product_attribute_exclude(self, attribute_number, value):
        key = 'product__attribute_%s' % attribute_number
        self.query = self.query.exclude(**{key: value})

    def set_product_attribute_exclude_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'product__attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.query = self.query.exclude(**{key: field_value})

    def set_exclude(self, lookup, value):
        self.excludes[lookup] = value

    def set_filter(self, lookup, value):
        self.filters[lookup] = value

    def set_search(self, s):
        # create default list
        qlist = [
            Q(name__icontains=s),
            Q(ecommerce_description__icontains=s),
            Q(product__name__icontains=s),
        ]

        # add config items
        config = settings.SALEBOX.get('SEARCH')
        if config:
            for qstr in config:
                qlist.append(Q((qstr, s)))

        # update query
        self.query = self.query.filter(reduce(operator.or_, qlist))

    def set_variant_attribute_include(self, attribute_number, value):
        key = 'attribute_%s' % attribute_number
        self.query = self.query.filter(**{key: value})

    def set_variant_attribute_include_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.query = self.query.filter(**{key: field_value})

    def set_variant_attribute_exclude(self, attribute_number, value):
        key = 'attribute_%s' % attribute_number
        self.query = self.query.exclude(**{key: value})

    def set_variant_attribute_exclude_keyvalue(
            self,
            attribute_number,
            field_name,
            field_value,
            field_modifier=None
        ):
        key = 'attribute_%s__%s' % (attribute_number, field_name)
        if field_modifier is not None:
            key = '%s__%s' % (key, field_modifier)
        self.query = self.query.exclude(**{key: field_value})

    def _retrieve_results(self, variant_ids, preserve_order=False):
        qs = []
        if len(variant_ids) > 0:
            qs = ProductVariant \
                    .objects \
                    .filter(id__in=variant_ids) \
                    .select_related('product', 'product__category')

            # prefetch attributes
            if len(self.prefetch_product_attributes) > 0:
                qs = qs.prefetch_related(*self.prefetch_product_attributes)
            if len(self.prefetch_variant_attributes) > 0:
                qs = qs.prefetch_related(*self.prefetch_variant_attributes)

            # price modifier: flat_discount
            if self.flat_discount > 0:
                ratio = 1 - (self.flat_discount / 100)
                qs = qs.annotate(
                    modified_price=F('price') * ratio
                )

            # price modifier: flat_member_discount
            if self.flat_member_discount > 0:
                ratio = 1 - (self.flat_member_discount / 100)
                qs = qs.annotate(modified_price=Case(
                    When(
                        member_discount_applicable=True,
                        then=F('price') * ratio
                    ),
                    default=F('price')
                ))

            # add ordering
            if preserve_order:
                preserved = Case(*[
                    When(pk=pk, then=pos) for pos, pk in enumerate(variant_ids)
                ])
                qs = qs.order_by(preserved)
            elif len(self.order) > 0:
                if (
                    self.flat_discount > 0 or
                    self.flat_member_discount > 0
                ):
                    self.order = [
                        o.replace('sale_price', 'modified_price')
                        for o in self.order
                    ]
                qs = qs.order_by(*self.order)

            # add offset / limit
            if self.max_result_count:
                qs = qs[self.offset:min(self.limit, self.max_result_count)]
            else:
                qs = qs[self.offset:self.limit]

            # modify results
            for o in qs:
                # flat discount modifiers
                try:
                    if o.modified_price:
                        o.sale_price = o.modified_price
                        del o.modified_price
                except:
                    pass

        return qs

    def _retrieve_variant_ids(self, do_exclude=False):
        self.set_active_status()

        if do_exclude:
            # this is complex:
            # if we have two sibling variants and one is sold out and one isn't, we need to
            # make sure we show the in-stock option in the list. Showing 'sold out' for one
            # variant is misleading as there are other options available the customer may
            # want
            sql = """
                SELECT          pv.id
                FROM            saleboxdjango_productvariant AS pv
                INNER JOIN      saleboxdjango_product AS p ON pv.product_id = p.id
                WHERE           p.id IN (
                    SELECT          p.id
                    FROM            saleboxdjango_product AS p
                    INNER JOIN      saleboxdjango_productvariant AS pv ON pv.product_id = p.id
                    WHERE           p.active_flag = true
                    AND             p.inventory_flag = True
                    AND             pv.active_flag = true
                    AND             pv.available_on_ecom = true
                    GROUP BY        p.id
                    HAVING          COUNT(pv) > 1
                    AND             SUM(pv.stock_total) > 0
                )
                AND             stock_total <= 0
                AND             p.inventory_flag = True
                ORDER BY        id
            """
            with connection.cursor() as cursor:
                cursor.execute(sql)
                self.set_exclude_productvariant_ids([row[0] for row in cursor.fetchall()])

        # return variant IDs
        if len(self.exclude_productvariant_ids) > 0:
            self.query = self.query.exclude(id__in=self.exclude_productvariant_ids)
        return list(self.query)

    def _sort_by_related_items(self, variant_ids):
        # generate 'order by' subquery
        subquery = []
        if self.order_related['string']['string']:
            subquery.append('SIMILARITY(pv.name, \'%s\') * %s' % (
                self.order_related['string']['string'].replace("'", "''"),
                self.order_related['string']['weight']
            ))

        for f in self.order_related['fields']:
            if f['type'] == 'category':
                subquery.append('(CASE WHEN p.category_id = %s THEN %s ELSE 0 END)' % (
                    f['values'][0],
                    f.get('weight', 1)
                ))
            if f['type'] == 'product':
                subquery.append('((SELECT COUNT(*) FROM saleboxdjango_product_attribute_%s WHERE product_id = p.id AND attributeitem_id IN (%s)) * %s)' % (
                    f['id'],
                    ','.join(str(i) for i in f['values']),
                    f.get('weight', 1)
                ))
            if f['type'] == 'variant':
                subquery.append('((SELECT COUNT(*) FROM saleboxdjango_productvariant_attribute_%s WHERE productvariant_id = p.id AND attributeitem_id IN (%s)) * %s)' % (
                    f['id'],
                    ','.join(str(i) for i in f['values']),
                    f.get('weight', 1)
                ))

        # create main sql
        sql = """
            SELECT              pv.id
            FROM                saleboxdjango_productvariant AS pv
            INNER JOIN          saleboxdjango_product AS p ON pv.product_id = p.id
            WHERE               pv.id IN (%s)
            [CATEGORY_IDS]
            ORDER BY            (%s) DESC, name_sorted
        """ % (
            ','.join(str(i) for i in variant_ids),
            ' + '.join(subquery)
        )

        # filter by category IDs if applicable
        if len(self.order_related['category_ids']) > 0:
            sql = sql.replace('[CATEGORY_IDS]', 'AND p.category_id IN (%s)' % ','.join([str(i) for i in self.order_related['category_ids']]))
        else:
            sql = sql.replace('[CATEGORY_IDS]', '')

        # do query
        with connection.cursor() as cursor:
            cursor.execute(sql)
            variant_ids = [row[0] for row in cursor.fetchall()]

        #
        return variant_ids


def translate_path(path):
    path = path.strip('/')
    condensed_path = re.sub('/+', '/', path)

    # 404 on double slashes
    if condensed_path != path.replace('//', '/'):
        raise Http404()

    # create empty object
    o = {
        'page_number': 1,
        'path': condensed_path,
        'path_list': condensed_path.split('/')
    }

    # strip empty path
    if o['path'] == '':
        o['path_list'] = []
    else:
        try:
            # is the last url segment a page number?
            o['page_number'] = int(o['path_list'][-1])
            del o['path_list'][-1]
            o['path'] = '/'.join(o['path_list'])

            # 404 if invalid page number
            if o['page_number'] < 1:
                raise Http404()
        except:
            pass

    #
    return o
